from .prim import load_primitives

class Context(object):
    def __init__(self, form, env, parent=None):
        self.form = form
        self.env = env
        self.vm = env.vm
        self.parent = parent

        self.ip = 0
        if self.form is not None:
            self.bytecode = form.bytecode
        else:
            self.bytecode = []
        self.stack = []

    def clone(self):
        "Make a clone of the context object."
        ctx = Context(self.form, self.env, self.parent)
        ctx.ip = self.ip
        ctx.stack = list(self.stack)
        return ctx

    def push(self, val):
        "Push a value onto the stack."
        self.stack.append(val)
    def pop(self):
        "Pop a value from the top of the stack."
        return self.stack.pop()
    def pop_n(self, n):
        "Remove n values from the top of the stack."
        del self.stack[-n:]
    def top(self, idx=1):
        "Get a value from the stack."
        return self.stack[-idx]
    def insert(self, idx, val):
        """\
        Insert value into particular position.
          [1, 2, 3] => insert(-1, 5) => [1, 2, 5, 3]
        """
        self.stack.insert(idx, val)
    def insert_n(self, idx, val):
        """\
        Insert a list of values into particular position.
          [1, 2] => insert(-1, [7, 8]) => [1, 7, 8, 2]
        """
        self.stack[idx:idx] = val

    def __str__(self):
        return '<Context stack_size=%d, ip=%d>' % (len(self.stack), self.ip)



from types          import NoneType
                      
from ..types.symbol import Symbol as sym
from ..types.pair   import Pair as pair
from ..macro        import Macro, DynamicClosure, SymbolClosure
from ..form         import Form
                     
from ..errors       import CompileError
from ..errors       import SyntaxError

from .builder       import Builder

class Compiler(object):
    """\
    The compiler for skime. It compiles sexp to bytecode.
    """

    sym_begin = sym("begin")
    sym_define = sym("define")
    sym_set_x = sym("set!")
    sym_if = sym("if")
    sym_lambda = sym("lambda")
    sym_quote = sym("quote")
    sym_or = sym("or")
    sym_and = sym("and")
    sym_define_syntax = sym("define-syntax")
    sym_syntax_rules = sym("syntax-rules")
    sym_let = sym("let")
    sym_letrec = sym("letrec")
    sym_letstar = sym("let*")
    sym_do = sym("do")
    sym_cond = sym("cond")
    sym_call_cc = sym("call/cc")
    sym_call_cc2 = sym("call-with-current-continuation")
    
    def __init__(self):
        self.label_seed = 0

    def compile(self, sexp, env):
        bdr = Builder(env)

        self.generate_expr(bdr, sexp, keep=True, tail=False)

        form = bdr.generate()
        return form

    ########################################
    # Helper functions
    ########################################
    def calc_env_distance(self, ancestor, descendant):
        "Calculate the distance between ancestor and descendant."
        dist = 0
        while descendant != ancestor:
            if descendant is None:
                raise SyntaxError("Attempt calculate the distance between unrelated environments.")
            dist += 1
            descendant = descendant.parent
        return dist
        
    def get_macro(self, env, name):
        if not isinstance(name, sym):
            return None
        loc = env.lookup_location(name.name)
        if loc is None:
            return None
        val = loc.env.read_local(loc.idx)
        if isinstance(val, Macro):
            return val
        return None
    
    def self_evaluating(self, expr):
        for t in [int, long, complex, float, str, unicode, bool, NoneType]:
            if isinstance(expr, t):
                return True
        return False
        
    def next_label(self):
        self.label_seed += 1
        return "__lbl_%d" % self.label_seed

    def generate_body(self, bdr, body, keep=True, tail=False):
        "Generate a sequence of sexps."
        if body is None and keep:
            bdr.emit("push_nil")
            if tail:
                bdr.emit("ret")

        while body is not None:
            expr = body.first
            body = body.rest
            will_keep = keep and body is None
            self.generate_expr(bdr, expr, keep=will_keep, tail=will_keep and tail)

    def generate_expr(self, bdr, expr, keep=True, tail=False):
        """\
        Generate instructions for an expression.

        if keep == True, the value of the expression is kept on
        the stack, otherwise, it is popped or never pushed.

        if tail == True, a tail call or ret will be emitted. tail
        can never be true if keep is False.
        """
        mapping = {
            Compiler.sym_if: self.generate_if_expr,
            Compiler.sym_begin: self.generate_body,
            Compiler.sym_lambda: self.generate_lambda,
            Compiler.sym_define: self.generate_define,
            Compiler.sym_set_x: self.generate_set_x,
            Compiler.sym_quote: self.generate_quote,
            Compiler.sym_or: self.generate_or,
            Compiler.sym_and: self.generate_and,
            Compiler.sym_define_syntax: self.generate_define_syntax,
            Compiler.sym_let: self.generate_let,
            Compiler.sym_letrec: self.generate_letrec,
            Compiler.sym_letstar: self.generate_letstar,
            Compiler.sym_do: self.generate_do,
            Compiler.sym_cond: self.generate_cond,
            Compiler.sym_call_cc: self.generate_call_cc,
            Compiler.sym_call_cc2: self.generate_call_cc
            }
        if self.self_evaluating(expr):
            if keep:
                bdr.emit('push_literal', expr)
                if tail:
                    bdr.emit('ret')
        
        elif isinstance(expr, sym):
            if keep:
                bdr.emit_local("push", expr.name)
                if tail:
                    bdr.emit('ret')

        elif isinstance(expr, pair):
            routine = mapping.get(expr.first)
            if routine is not None:
                routine(bdr, expr.rest, keep=keep, tail=tail)
            else:
                argc = 0
                macro = self.get_macro(bdr.env, expr.first)
                transform_env = bdr.env
                if macro is not None:
                    expr, dc_list = macro.transform(transform_env, expr)

                    transform_env = macro.lexical_parent
                    form_bdr = Builder(transform_env)
                    self.generate_expr(form_bdr, expr, keep=True, tail=False)
                    macro_closure = DynamicClosure(transform_env, expr)
                    macro_closure.form = form_bdr.generate()
                    bdr.emit('push_literal', macro_closure)

                    dist = self.calc_env_distance(macro.lexical_parent, bdr.env)
                    if dist == 0:
                        bdr.emit('fix_lexical')
                    else:
                        bdr.emit('fix_lexical_depth', dist)

                    # fix lexical parent of dynamic closures
                    for dc in dc_list:
                        bdr.emit('push_literal', dc)
                        bdr.emit('fix_lexical_pop')
                        form_bdr = Builder(bdr.env)
                        self.generate_expr(form_bdr, dc.expression, keep=True, tail=False)
                        dc.form = form_bdr.generate()

                    bdr.emit('dynamic_eval')
                    if not keep:
                        bdr.emit('pop')
                    elif tail:
                        bdr.emit('ret')
                        
                else:
                    arg  = expr.rest
                    while isinstance(arg, pair):
                        self.generate_expr(bdr, arg.first, keep=True, tail=False)
                        arg = arg.rest
                        argc += 1
                    self.generate_expr(bdr, expr.first, keep=True, tail=False)
                    
                    if tail:
                        bdr.emit('tail_call', argc)
                    else:
                        bdr.emit('call', argc)
                        if not keep:
                            bdr.emit('pop')

        elif isinstance(expr, DynamicClosure):
            bdr.emit('push_literal', expr)
            bdr.emit('dynamic_eval')
            if not keep:
                bdr.emit('pop')
            elif tail:
                bdr.emit('ret')
            
        else:
            raise CompileError("Expecting atom or list, but got %s" % expr)

    def generate_if_expr(self, bdr, expr, keep=True, tail=False):
        if expr is None:
            raise SyntaxError("Missing condition expression in 'if'")
            
        cond = expr.first
        expthen = expr.rest
        if expthen is None:
            raise SyntaxError("Missing 'then' expression in 'if'")
        expthen = expthen.first

        expelse = expr.rest.rest
        if expelse is not None:
            if expelse.rest is not None:
                raise SyntaxError("Extra expression in 'if'")
            expelse = expelse.first

        self.generate_expr(bdr, cond, keep=True, tail=False)

        if keep is True:
            lbl_then = self.next_label()
            lbl_end = self.next_label()
            bdr.emit('goto_if_not_false', lbl_then)
            if expelse is None:
                bdr.emit('push_nil')
                if tail:
                    bdr.emit('ret')
            else:
                self.generate_expr(bdr, expelse, keep=True, tail=tail)
            if not tail:
                bdr.emit('goto', lbl_end)
            bdr.def_label(lbl_then)
            self.generate_expr(bdr, expthen, keep=True, tail=tail)
            bdr.def_label(lbl_end)
        else:
            if expelse is None:
                lbl_end = self.next_label()
                bdr.emit('goto_if_false', lbl_end)
                self.generate_expr(bdr, expthen, keep=False, tail=False)
                bdr.def_label(lbl_end)
            else:
                lbl_then = self.next_label()
                lbl_end = self.next_label()
                bdr.emit('goto_if_not_false', lbl_then)
                self.generate_expr(bdr, expelse, keep=False, tail=False)
                bdr.emit('goto', lbl_end)
                bdr.def_label(lbl_then)
                self.generate_expr(bdr, expthen, keep=False, tail=False)
                bdr.def_label(lbl_end)

    def filter_sc(self, expr):
        if isinstance(expr, SymbolClosure):
            return expr.expression
        elif isinstance(expr, sym):
            return expr
        raise SyntaxError("Expecting symbol, but got %s" % expr)
    
    def generate_lambda(self, base_builder, expr, keep=True, tail=False):
        if keep is not True:
            return  # lambda expression has no side-effect
        try:
            arglst = expr.first
            body = expr.rest

            if isinstance(arglst, pair):
                args = []
                while isinstance(arglst, pair):
                    args.append(self.filter_sc(arglst.first).name)
                    arglst = arglst.rest
                if arglst is None:
                    rest_arg = False
                else:
                    args.append(self.filter_sc(arglst).name)
                    rest_arg = True
            elif arglst is None:
                rest_arg = False
                args = []
            else:
                rest_arg = True
                args = [self.filter_sc(arglst).name]

            bdr = base_builder.push_proc(args=args, rest_arg=rest_arg)
            self.generate_body(bdr, body, keep=True, tail=True)
            base_builder.emit("fix_lexical")
            
            if tail:
                base_builder.emit('ret')

        except AttributeError, e:
            raise SyntaxError("Broken lambda expression: "+e.message)

    def generate_let(self, bdr, expr, keep=True, tail=False):
        """\
        (let ((var1 val1) (var2 val2)) expr1 expr2)
                          ||
                         _||_
                          \/
        ((lambda (var1 var2) expr1 expr2) val1 val2)
        """
        if not isinstance(expr, pair):
            raise SyntaxError("Invalid let expression")
        bindings = expr.first
        param = []
        args = []

        if isinstance(bindings, pair):
            while isinstance(bindings, pair):
                binding = bindings.first
                if not isinstance(binding, pair) or \
                   not isinstance(binding.rest, pair):
                    raise SyntaxError("Invalid binding for let expression: %s" % binding)
                param.append(self.filter_sc(binding.first).name)
                args.append(binding.rest.first)
                bindings = bindings.rest
        elif bindings is not None:
            raise SyntaxError("Invalid let expression: expecting bindings, but got %s" % bindings)

        for x in args:
            self.generate_expr(bdr, x, keep=True, tail=False)

        lambda_bdr = bdr.push_proc(args=param, rest_arg=False)
        self.generate_body(lambda_bdr, expr.rest, keep=True, tail=True)
        bdr.emit('fix_lexical')

        argc = len(args) 
        if tail:
            bdr.emit('tail_call', argc)
        else:
            bdr.emit('call', argc)
            if not keep:
                bdr.emit('pop')

    def generate_letrec(self, bdr, expr, keep=True, tail=False):
        if not isinstance(expr, pair):
            raise SyntaxError("Invalid letrec expression")
        
        bindings = expr.first
        body = expr.rest

        lambda_bdr = bdr.push_proc()

        # letrec will evaluate the init forms in the new env
        names = []
        vals = []
        
        while isinstance(bindings, pair):
            binding = bindings.first
            if not isinstance(binding, pair) or \
               not isinstance(binding.first, sym) or \
               not isinstance(binding.rest, pair):
                raise SyntaxError("Invalid binding for letrec expression: %s" % binding)
            name = self.filter_sc(binding.first).name
            val = binding.rest.first
            lambda_bdr.def_local(name)

            names.append(name)
            vals.append(val)

            bindings = bindings.rest
        
        if bindings is not None:
            raise SyntaxError("Invalid bindings for letrec expression: %s" % bindings)

        for i in range(len(names)):
            self.generate_expr(lambda_bdr, vals[i], keep=True, tail=False)
            lambda_bdr.emit_local('set', names[i])

        self.generate_body(lambda_bdr, body, keep=True, tail=True)
        bdr.emit('fix_lexical')

        if tail:
            bdr.emit('tail_call', 0)
        else:
            bdr.emit('call', 0)
            if not keep:
                bdr.emit('pop')

    def generate_letstar(self, bdr, expr, keep=True, tail=False):
        if not isinstance(expr, pair):
            raise SyntaxError("Invalid letrec expression")
        
        bindings = expr.first
        body = expr.rest

        lambda_bdr = bdr.push_proc()

        # let* will evaluate the init forms in the new env sequencially
        names = []
        vals = []
        
        while isinstance(bindings, pair):
            binding = bindings.first
            if not isinstance(binding, pair) or \
               not isinstance(binding.first, sym) or \
               not isinstance(binding.rest, pair):
                raise SyntaxError("Invalid binding for let* expression: %s" % binding)
            name = self.filter_sc(binding.first).name
            val = binding.rest.first

            names.append(name)
            vals.append(val)

            bindings = bindings.rest
        
        if bindings is not None:
            raise SyntaxError("Invalid bindings for let* expression: %s" % bindings)

        for i in range(len(names)):
            lambda_bdr.def_local(names[i])
            self.generate_expr(lambda_bdr, vals[i], keep=True, tail=False)
            lambda_bdr.emit_local('set', names[i])

        self.generate_body(lambda_bdr, body, keep=True, tail=True)
        bdr.emit('fix_lexical')

        if tail:
            bdr.emit('tail_call', 0)
        else:
            bdr.emit('call', 0)
            if not keep:
                bdr.emit('pop')
        
        
    def generate_define(self, bdr, expr, keep=True, tail=False):
        if expr is None:
            raise SyntaxError("Empty define expression")
        var = expr.first

        # SymbolClosure now has no effect, only as normal symbol
        if isinstance(var, SymbolClosure):
            var = var.expression
        
        if isinstance(var, pair):
            gen = self.generate_lambda
            val = pair(var.rest, expr.rest)
            var = var.first
        elif isinstance(var, sym):
            gen = self.generate_expr
            val = expr.rest
            if val is None:
                raise SyntaxError("Missing value for defined variable")
            if val.rest is not None:
                raise SyntaxError("Extra expressions in 'define'")
            val = val.first
        else:
            raise SyntaxError("Invalid define expression")

        # first define local, then generate value. This allow
        # recursive function to be compiled properly.
        bdr.def_local(var.name)
        gen(bdr, val, keep=True, tail=False)
        if keep is True:
            bdr.emit('dup')
        bdr.emit_local('set', var.name)
        if tail:
            bdr.emit('ret')

    def generate_set_x(self, bdr, expr, keep=True, tail=False):
        if expr is None:
            raise SyntaxError("Empty set! expression")
        var = expr.first
        val = expr.rest

        if val is None:
            raise SyntaxError("Missing value for set! expression")
        if val.rest is not None:
            raise SyntaxError("Extra expressions in 'set!'")
        val = val.first

        self.generate_expr(bdr, val, keep=True, tail=False)
        if keep:
            bdr.emit('dup')

        if isinstance(var, sym):
            bdr.emit_local('set', var.name)
        elif isinstance(var, SymbolClosure):
            bdr.emit('push_literal', var)
            bdr.emit_local('set', var.expression.name, var.lexical_parent)
        else:
            raise SyntaxError("Invalid set! expression, expecting symbol")

        if tail:
            bdr.emit('ret')

    def generate_quote(self, bdr, expr, keep=True, tail=False):
        expr = expr.first
        if isinstance(expr, DynamicClosure):
            expr = expr.expression
        if keep:
            bdr.emit('push_literal', expr)
            if tail:
                bdr.emit('ret')

    def generate_or(self, bdr, expr, keep=True, tail=False):
        lbl_end = self.next_label()
        expr_generated = False
        while isinstance(expr, pair):
            el = expr.first
            expr = expr.rest
            # 'False' literal in 'or' expression can be silently
            # ignored
            if el is not False:
                expr_generated = True
                self.generate_expr(bdr, el, keep=True, tail=False)
                if keep:
                    bdr.emit('dup')
                bdr.emit('goto_if_not_false', lbl_end)
                if keep:
                    if expr is not None:
                        bdr.emit('pop')
        if expr is not None:
            raise SyntaxError("Invalid element in or expression: %s" % expr)
        if keep:
            if not expr_generated:
                bdr.emit('push_false')
            if tail:
                bdr.emit('ret')
                
        bdr.def_label(lbl_end)

    def generate_and(self, bdr, expr, keep=True, tail=False):
        lbl_end = self.next_label()
        expr_generated = False
        while isinstance(expr, pair):
            el = expr.first
            expr = expr.rest
            # 'True' literal in 'and' expression can be silently
            # ignored
            if el is not True:
                expr_generated = True
                self.generate_expr(bdr, el, keep=True, tail=False)
                if keep:
                    bdr.emit('dup')
                bdr.emit('goto_if_false', lbl_end)
                if keep:
                    if expr is not None:
                        bdr.emit('pop')
        if expr is not None:
            raise SyntaxError("Invalid element in or expression: %s" % expr)
        if keep:
            if not expr_generated:
                bdr.emit('push_true')
            if tail:
                bdr.emit('ret')
                
        bdr.def_label(lbl_end)

    def generate_define_syntax(self, bdr, expr, keep=True, tail=False):
        if not isinstance(expr, pair):
            raise SyntaxError("Invalid define-syntax expression, expecting macro keyword")
        name = expr.first
        if not isinstance(name, sym):
            raise SyntaxError("Expecting macro keyword as a symbol, but got %s" % name)
        expr = expr.rest
        if not isinstance(expr, pair) or \
               not isinstance(expr.first, pair) or \
               Compiler.sym_syntax_rules != expr.first.first:
            raise SyntaxError("Expecting syntax-rules, but got %s" % expr.first)
        if expr.rest is not None:
            raise SyntaxError("Extra expressions in define-syntax: %s" % expr.rest)
        
        # define local before constructing the macro, so that recursive macro
        # can be supported
        idx = bdr.def_local(name.name)
        macro = Macro(bdr.env, expr.first.rest)
        bdr.env.assign_local(idx, macro)

        if keep:
            # macro object is generally not available at runtime, the value of
            # 'define-syntax' expression is None
            bdr.emit('push_nil')
            if tail:
                bdr.emit('ret')

    def generate_do(self, bdr, expr, keep=True, tail=False):
        if not isinstance(expr, pair):
            raise SyntaxError("Invalid do expression")
        init_spec = expr.first

        expr = expr.rest
        if not isinstance(expr, pair) or \
           not isinstance(expr.first, pair):
            raise SyntaxError("Invalid do expression, expecting (<test> <result>)")

        test_expr = expr.first.first
        result_expr = expr.first.rest
        body = expr.rest

        variables = []
        init_vals = []
        steps = []
        while isinstance(init_spec, pair):
            spec = init_spec.first
            if not isinstance(spec, pair) or \
               not isinstance(spec.rest, pair):
                raise SyntaxError("Invalid init spec for do expression: %s" % spec)
            var = self.filter_sc(spec.first)
            if not isinstance(var, sym):
                raise SyntaxError("Invalid init spec for do expression: %s" % spec)
            variables.append(var.name)
            init_vals.append(spec.rest.first)

            if isinstance(spec.rest.rest, pair):
                steps.append(spec.rest.rest.first)
            else:
                steps.append(None)

            init_spec = init_spec.rest
        if init_spec is not None:
            raise SyntaxError("Invalid init specs for do expression")
            
        for val in init_vals:
            self.generate_expr(bdr, val, keep=True, tail=False)

        lam_bdr = bdr.push_proc(args=variables, rest_arg=False)
        lbl_test = self.next_label()
        lbl_end = self.next_label()

        lam_bdr.def_label(lbl_test)
        self.generate_expr(lam_bdr, test_expr, keep=True, tail=False)
        lam_bdr.emit('goto_if_not_false', lbl_end)
        self.generate_body(lam_bdr, body, keep=False, tail=False)

        for i in range(len(steps)):
            if steps[i] is not None:
                self.generate_expr(lam_bdr, steps[i], keep=True, tail=False)
        for i in range(len(steps)-1, -1, -1):
            if steps[i] is not None:
                lam_bdr.emit_local('set', variables[i])

        lam_bdr.emit('goto', lbl_test)
        lam_bdr.def_label(lbl_end)
        self.generate_body(lam_bdr, result_expr, keep=True, tail=True)

        bdr.emit('fix_lexical')

        if tail:
            bdr.emit('tail_call', len(variables))
        else:
            bdr.emit('call', len(variables))
            if not keep:
                bdr.emit('pop')

    def generate_cond(self, bdr, expr, keep=True, tail=False):
        if not isinstance(expr, pair):
            raise SyntaxError("Empty cond expression")

        lbl_end = self.next_label()
        lbl_next = self.next_label()
        first = True

        while isinstance(expr, pair):
            bdr.def_label(lbl_next)
            lbl_next = self.next_label()

            if not first:
                bdr.emit('pop')
                first = False
            
            cond_expr = expr.first
            if not isinstance(cond_expr, pair):
                raise SyntaxError("Invalid cond clause: %s" % cond_expr)
            pred = cond_expr.first
            body = cond_expr.rest

            expr = expr.rest

            if pred == sym('else'):
                if body is None:
                    bdr.emit('push_true')
                else:
                    if not isinstance(body, pair):
                        raise SyntaxError("Invalid cond clause: %s" % cond_expr)
                    if body.first == sym('=>'):
                        if not isinstance(body.rest, pair):
                            raise SyntaxError("Invalid cond clause, expecting expression after =>")
                        bdr.emit('push_true')
                        self.generate_expr(bdr, body.rest.first, keep=True, tail=False)
                        bdr.emit('call', 1)
                    else:
                        self.generate_body(bdr, body, keep=True, tail=False)
                break
            
            else:
                self.generate_expr(bdr, pred, keep=True, tail=False)
                if body is None:
                    bdr.emit('dup')
                    bdr.emit('goto_if_false', lbl_next)
                else:
                    if not isinstance(body, pair):
                        raise SyntaxError("Invalid cond clause: %s" % cond_expr)
                    if body.first == sym('=>'):
                        if not isinstance(body.rest, pair):
                            raise SyntaxError("Invalid cond clause, expecting expression after =>")
                        bdr.emit('dup')
                        bdr.emit('goto_if_false', lbl_next)
                        self.generate_expr(bdr, body.rest.first, keep=True, tail=False)
                        bdr.emit('call', 1)
                    else:
                        bdr.emit('goto_if_false', lbl_next)
                        self.generate_body(bdr, body, keep=True, tail=False)
                bdr.emit('goto', lbl_end)
                    
        if expr is not None:
            raise SyntaxError("Extra garbage expression in cond expression: %s" % expr)
        
        bdr.def_label(lbl_next)
        bdr.emit('push_nil')
        bdr.def_label(lbl_end)

        if not keep:
            bdr.emit('pop')
        if tail:
            bdr.emit('ret')

    def generate_call_cc(self, bdr, expr, keep=True, tail=False):
        if not isinstance(expr, pair):
            raise SyntaxError("Empty call/cc expression")
        if expr.rest is not None:
            raise SyntaxError("call/cc only takes one argument, but got extra %s" % expr.rest)

        lam = expr.first
        self.generate_expr(bdr, lam, keep=True, tail=False)
        bdr.emit('call_cc')
        if tail:
            bdr.emit('ret')
        if not keep:
            bdr.emit('pop')



import re
import yaml

def gen_tags(tags):
    stmts =  []
    for tag, i in zip(tags, range(len(tags))):
        stmts.append('TAG_%-12s = %d' % (tag.upper(), 2**i))
    return '\n'.join(stmts)

def gen_actions(instructions):
    def gen_action(insn):
        func = "def op_%s(ctx):\n" % insn['name']

        env = {
            'insn_len' : 1 + len(insn['operands'])
            }
        code = insn['code']
        if not 'ctrl_flow' in insn['tags']:
            code += 'ctx.ip += $(insn_len)\n'
            
        code = process_tmpl(code, env)
        code = re.sub(re.compile('^', re.MULTILINE), '    ', code)
        
        return func + code

    return '\n'.join([gen_action(insn)
                      for insn in instructions])

def gen_action_table(instructions):
    return 'INSN_ACTION = [\n' + \
           ',\n'.join(['    op_' + insn['name']
                       for insn in instructions]) + \
           '\n]\n'


def gen_tags_table(instructions):
    def gen_tag(insn):
        if len(insn['tags']) == 0:
            return '0'
        else:
            return ' | '.join(['TAG_%s' % tag.upper()
                               for tag in insn['tags']])

    return 'INSN_TAGS = [\n' + \
           ',\n'.join(['    ' + gen_tag(insn)
                       for insn in instructions]) + \
           '\n]\n'

def gen_insn_table(instructions):
    def gen_insn(i, insn):
        return "Instruction(" + str(i) + ",\n" + \
               ",\n".join(["            " + insn[key].__repr__()
                           for key in ['name', 'tags', 'desc', 'operands',
                                       'stack_before', 'stack_after', 'code']]) + \
               ")"

    insns = zip(range(len(instructions)), instructions)
    
    return "INSTRUCTIONS = [\n" + \
           ",\n".join([gen_insn(i, insn)
                      for i, insn in insns]) + \
           "]\n\nINSN_MAP = {\n" + \
           ",\n".join(['    ' + insn['name'].__repr__() + ' : INSTRUCTIONS[%d]' % i
                       for i, insn in insns]) + \
           "\n}\n"
    

def process_tmpl(tmpl, env):
    """\
    Process template. Special variables like $(key) in the template
    will be replaced by the value found in env (env['key']).
    """
    PATTERN = re.compile(r"\$\(([^)]+)\)")
    return re.sub(PATTERN, lambda m: str(env[m.group(1)]), tmpl)
    
    
TMPL_INSNS = """\
# Don't edit this file. This is generated by iset_gen.py

from .ctx        import Context
from .call_cc    import Continuation
from .proc       import Procedure
from .prim       import Primitive
from .types.pair import Pair
from .errors     import WrongArgType

$(tags)

$(actions)

$(action_table)

$(tags_table)


def has_tag(opcode, tag):
    return INSN_TAGS[opcode] & tag == tag

def get_param(ctx, n):
    return ctx.bytecode[ctx.ip+n]

def run(ctx):
    while ctx.ip < len(ctx.bytecode):
        opcode = ctx.bytecode[ctx.ip]
        nctx = INSN_ACTION[opcode](ctx)
        if has_tag(opcode, TAG_CTX_SWITCH):
            ctx = nctx
    return ctx.pop()

def make_call(ctx, argc, tail=False):
    proc = ctx.pop()
    if tail:
        parent = ctx.parent
    else:
        parent = ctx

    if isinstance(proc, Procedure):
        proc.check_arity(argc)
        nctx = Context(proc, proc.env.dup(), parent)

        for i in range(proc.fixed_argc):
            nctx.env.assign_local(i, ctx.top(idx=argc-i))
        if proc.fixed_argc != proc.argc:
            rest = None
            for i in range(argc-proc.fixed_argc):
                rest = Pair(ctx.top(idx=i+1), rest)
            nctx.env.assign_local(proc.fixed_argc, rest)
        ctx.pop_n(argc)

    elif isinstance(proc, Primitive):
        proc.check_arity(argc)
        args = range(argc)
        for i in range(argc):
            args[i] = ctx.top(idx=argc-i)
        ctx.pop_n(argc)

        nctx = parent
        nctx.push(proc.call(ctx.vm, *args))

    elif isinstance(proc, Continuation):
        if argc > 1:
            raise WrongArgNumber("Continuation only accept 1 argument")
        nctx = proc.ctx.clone()
        if argc == 1:
            nctx.push(ctx.pop())
        else:
            nctx.push(None)
        nctx.parent = ctx.parent

    else:
        raise WrongArgType("Not a skime callable: %s" % proc)

    return nctx
"""

TMPL_ISET = """\
# Don't edit this file. This is generated by iset_gen.py

class Instruction(object):
    __slots__ = ('opcode',
                 'name',
                 'tags',
                 'desc',
                 'operands',
                 'stack_before',
                 'stack_after',
                 'code')
    def __init__(self, opcode, name, tags, desc, operands,
                 stack_before, stack_after, code):
        self.opcode = opcode
        self.name = name
        self.tags = tags
        self.desc = desc
        self.operands = operands
        self.stack_before = stack_before
        self.stack_after = stack_after
        self.code = code

    def length_get(self):
        return len(self.operands)+1
    def length_set(self):
        raise AttributeError, 'length attribute is read only'
    length = property(length_get, length_set, 'length of the instruction')

$(instruction_table)
"""

if __name__ == '__main__':
    iset = yaml.load(open("iset.yml").read())

    env = {
        'tags' : gen_tags(iset['tags']),
        'actions' : gen_actions(iset['instructions']),
        'instruction_table' : gen_insn_table(iset['instructions']),
        'action_table' : gen_action_table(iset['instructions']),
        'tags_table' : gen_tags_table(iset['instructions'])
        }

    py = open("iset.py", "w")
    py.write(process_tmpl(TMPL_ISET, env))
    py.close()

    py = open("insns.py", "w")
    py.write(process_tmpl(TMPL_INSNS, env))
    py.close()



from cStringIO        import StringIO

from .errors          import WrongArgNumber
from .compiler.disasm import disasm

class Procedure(object):
    def __init__(self, builder, bytecode):
        
        # The Environment created at compile time. A new
        # Environment will be created when the procedure
        # is called.
        #
        # self.env.parent (i.e. the lexical parent of the
        # env) might be an Environment created at compile
        # time. In that case, a 'make-lambda' instruction
        # will be emitted in the instruction sequence to
        # fix the parent at run time.
        self.env = builder.env
        
        self.bytecode = bytecode

        self.argc = len(builder.args)
        if builder.rest_arg:
            self.fixed_argc = self.argc-1
        else:
            self.fixed_argc = self.argc

        self.literals = list(builder.literals)

    def lexical_parent_get(self):
        return self.env.parent
    def lexical_parent_set(self, parent):
        self.env.parent = parent
    lexical_parent = property(lexical_parent_get, lexical_parent_set)

    def check_arity(self, argc):
        if self.fixed_argc == self.argc:
            if argc != self.argc:
                raise WrongArgNumber("Expecting %d arguments, but got %d" %
                                     (self.argc, argc))
        else:
            if argc < self.fixed_argc:
                raise WrongArgNumber("Expecting at least %d arguments, but got %d" %
                                     (self.fixed_argc, argc))
                
    def disasm(self):
        "Show the disassemble of the instructions of the proc. Useful for debug."
        io = StringIO()
        io.write('='*60)
        io.write('\n')
        io.write('Diasassemble of proc at %X\n' % id(self))
        
        io.write('arguments: ')
        args = [self.env.get_name(i) for i in range(self.argc)]
        if self.fixed_argc != self.argc:
            args[-1] = '*'+args[-1]
        io.write(', '.join(args))
        io.write('\n')

        io.write("literals:\n")
        for i in range(len(self.literals)):
            io.write("%4d: %s\n" % (i, self.literals[i]))

        io.write("\ninstructions:\n")
        io.write('-'*50)
        io.write('\n')

        disasm(io, self)

        content = io.getvalue()
        io.close()

        return content



from cStringIO        import StringIO

from .errors          import MiscError
from .env             import Environment
from .compiler.disasm import disasm
from .insns           import run
from .ctx             import Context

class Form(object):
    """\
    A Form is the result of compiling an expression.
    """
    def __init__(self, builder, bytecode):
        # The environment under which the form is compiled
        # This is a reference to an Environment that might
        # be shared by many forms. This is only used in
        # debugging to find the local variable names used
        # by this form.
        self.env = builder.env
        
        # The bytecode of this form
        self.bytecode = bytecode

        # The literals used in bytecode
        self.literals = builder.literals

    def eval(self, env, vm):
        "Eval the form under env and vm."
        ctx = Context(self, env, vm.ctx)
        return run(ctx)

    def disasm(self):
        "Show the disassemble of the instructions of the form. Useful for debug."
        io = StringIO()
        io.write('='*60)
        io.write('\n')
        io.write('Diasassemble of form at %X\n' % id(self))

        io.write("literals:\n")
        for i in range(len(self.literals)):
            io.write("%4d: %s\n" % (i, self.literals[i]))

        io.write("\ninstructions:\n")
        io.write('-'*50)
        io.write('\n')

        disasm(io, self)

        content = io.getvalue()
        io.close()

        return content



import weakref

class Symbol(object):
    symbols = weakref.WeakValueDictionary({})

    def __new__(cls, name):
        """\
        Get the interned symbol of name. If no found, create
        a new interned symbol.
        """
        if cls.symbols.has_key(name):
            return cls.symbols[name]

        sym = object.__new__(cls)
        sym._name = name
        cls.symbols[name] = sym
        return sym

    def __eq__(self, other):
        return self is other
    def __ne__(self, other):
        return not self.__eq__(other)

    def get_name(self):
        return self._name
    def set_name(self):
        raise AttributeError("Can't modify name of a symbol.")
    name = property(get_name, set_name)

    def __str__(self):
        return self.name
    def __repr__(self):
        return "<symbol %s>" % self._name.__repr__()



from .errors import WrongArgNumber

class Continuation(object):
    def __init__(self, ctx, ip_displacement, n_pop):
        self.ctx = ctx.clone()
        self.ctx.ip += ip_displacement
        if n_pop > 0:
            self.ctx.pop_n(n_pop)

    def __str__(self):
        return '<Continuation ctx=%s>' % self.ctx



from ..types.symbol import Symbol as sym
from ..types.pair   import Pair as pair

from ..errors import ParseError

def parse(text, name="__unknown__"):
    "Parse a piece of text."
    return Parser(text, name).parse()

class Parser(object):
    "A simple recursive descent parser for Scheme."
    sym_quote = sym("quote")
    sym_quasiquote = sym("quasiquote")
    sym_unquote = sym("unquote")
    sym_unquote_slicing = sym("unquote-slicing")
    
    def __init__(self, text, name="__unknown__"):
        self.text = text
        self.name = name
        self.pos = 0
        self.line = 1

    def parse(self):
        "Parse the text and return a sexp."
        expr = self.parse_expr()
        self.skip_all()
        if self.more():
            self.report_error("Expecting end of code, but more code is got")
        return expr


    def parse_expr(self):
        def parse_pound():
            if self.peak(idx=1) == 't':
                self.pop(n=2)
                return True
            if self.peak(idx=1) == 'f':
                self.pop(n=2)
                return False
            if self.peak(idx=1) == '(':
                return self.parse_vector()
        def parse_number_or_symbol():
            if self.isdigit(self.peak(idx=1)):
                return self.parse_number()
            return self.parse_symbol()
            
        mapping = {
            '#' : parse_pound,
            '(' : self.parse_list,
            "'" : self.parse_quote,
            '`' : self.parse_quote,
            ',' : self.parse_unquote,
            '+' : parse_number_or_symbol,
            '-' : parse_number_or_symbol,
            '"' : self.parse_string
            }

        self.skip_all()

        if not self.more():
            raise ParseError("Nothing to be parsed.")
        ch = self.peak()
        routine = mapping.get(ch)
        
        if routine is not None:
            return routine()
        
        if self.isdigit(ch):
            return self.parse_number()
        return self.parse_symbol()


    def parse_number(self):
        sign1 = 1
        if self.eat('-'):
            sign1 = -1
        self.eat('+')

        num1 = self.parse_unum()
        if self.eat('/'):
            num2 = self.parse_unum()
            if num2 is None:
                self.report_error("Invalid number format, expecting denominator")
            num1 = float(num1)/num2
        if self.peak() in ['+', '-']:
            sign2 = 1
            if self.eat('-'):
                sign2 = -1
            self.eat('+')
            num2 = self.parse_unum()
            if num2 is None:
                num2 = 1
            if not self.eat('i'):
                self.report_error("Invalid number format, expecting 'i' for complex")
            if num2 != 0:
                num1 = num1 + sign2*num2*1j

        return sign1*num1

    def parse_unum(self):
        "Parse an unsigned number."
        isfloat = False
        pos1 = self.pos
        while self.isdigit(self.peak()):
            self.pop()
        if self.eat('.'):
            isfloat = True
        while self.isdigit(self.peak()):
            self.pop()
        pos2 = self.pos
        if pos2 == pos1:
            return None
        if isfloat:
            return float(self.text[pos1:pos2])
        else:
            return int(self.text[pos1:pos2])


    def parse_list(self):
        self.eat('(')
        elems = []
        while self.more():
            self.skip_all()
            if self.peak() == ')':
                elems.append(None)
                break
            if self.peak() == '.' and self.peak(idx=1) != '.':
                self.eat('.')
                elems.append(self.parse_expr())
                self.skip_all()
                break
            elems.append(self.parse_expr())
        if not self.eat(')'):
            self.report_error("Expecting right paren ')'.")
        first = elems.pop()
        for x in reversed(elems):
            first = pair(x, first)
        return first

    def parse_quote(self):
        if self.peak() == '\'':
            s = Parser.sym_quote
        else:
            s = Parser.sym_quasiquote
        self.pop()
        return pair(s, pair(self.parse_expr(), None))

    def parse_unquote(self):
        self.eat(',')
        if self.eat('@'):
            return pair(Parser.sym_unquote_slicing,
                        pair(self.parse_expr(), None))
        return pair(Parser.sym_unquote,
                    pair(self.parse_expr(), None))

    def parse_symbol(self):
        pos1 = self.pos
        self.pop()
        while self.more() and \
              not self.isspace(self.peak()) and \
              not self.peak() in ['\'', ')', '(', ',', '@']:
            self.pop()
        pos2 = self.pos
        return sym(self.text[pos1:pos2])

    def parse_string(self):
        mappings = {
            '"':'"',
            '\\':'\\',
            'n':'\n',
            't':'\t'
            }
            
        self.eat('"')
        strings = []
        pos1 = self.pos
        while self.more():
            if self.peak() == '"':
                break
            if self.peak() == '\\':
                self.pop()
                ch = self.peak()
                if ch in mappings:
                    strings.append(self.text[pos1:self.pos-1])
                    strings.append(mappings[ch])
                    self.pop()
                    pos1 = self.pos
            else:
                if self.peak() == '\n':
                    self.line += 1
                self.pop()
        strings.append(self.text[pos1:self.pos])
        if not self.eat('"'):
            report_error("Expecting '\"' to end a string.")
        return ''.join(strings)
                

    def parse_vector(self):
        pass

    def skip_all(self):
        "Skip all non-relevant characters."
        while True:
            self.skip_ws()
            if self.peak() == ';':
                self.skip_comment()
            else:
                break

    def skip_ws(self):
        "Skip whitespace."
        while self.more():
            if self.eat('\n'):
                self.line += 1
            elif self.isspace(self.peak()):
                self.pop()
            else:
                break

    def skip_comment(self):
        "Skip comment."
        while self.eat(';'):
            while self.more() and not self.eat('\n'):
                self.pop()
            self.line += 1

    def pop(self, n=1):
        "Increase self.pos by n."
        self.pos += n

    def more(self):
        "Whether we have more content to parse."
        return self.pos < len(self.text)
            
    def eat(self, char):
        "Try to eat a character."
        if self.peak() != char:
            return False
        self.pos += 1
        return True

    def peak(self, idx=0):
        "Get the character under self.pos + idx."
        if self.pos + idx < len(self.text):
            return self.text[self.pos + idx]
        return None

    def isdigit(self, ch):
        "Test whether ch is a digit."
        return ch is not None and ch.isdigit()

    def isspace(self, ch):
        "Test whether ch is whitespace."
        return ch is not None and ch.isspace()
        
    def report_error(self, msg):
        "Raise a ParserError with msg."
        raise ParseError("%s:%d %s" % (self.name, self.line, msg))



from array    import array

from ..iset   import INSN_MAP
from ..form   import Form
from ..proc   import Procedure
from ..env    import Environment
from ..errors import UnboundVariable

class Builder(object):
    "Builder is a helper of building the bytecode for a form."
    def __init__(self, env, result_t=Form):
        # The lexical environment where the form is compiled
        self.env = env
        # The type of the generate result
        self.result_t = result_t

        # The instruction stream
        self.stream = []
        # The program counter (instruction pointer)
        self.ip = 0
        # The lable name => ip mapping
        self.labels = {}
        # The literals
        self.literals = []

    def emit(self, insn_name, *args):
        "Emit an instruction"
        insn = INSN_MAP.get(insn_name)
        if insn is None:
            raise TypeError, "No such instruction: %s" % insn_name
        if insn.length != 1+len(args):
            raise TypeError, \
                  "INSTRUCTION %s expects %d parameters, but %d given" % \
                  (insn_name, insn.length-1, len(args))

        if insn_name == 'push_literal':
            lit = args[0]
            # NOTE True == 1, False == 0 in Python, so "is True/False"
            # test should be put before "== 0/1" test.
            if lit is True:
                insn_name = 'push_true'
                args = ()
            elif lit is False:
                insn_name = 'push_false'
                args = ()
            elif lit == 0 and isinstance(lit, int):
                insn_name = 'push_0'
                args = ()
            elif lit == 1 and isinstance(lit, int):
                insn_name = 'push_1'
                args = ()
            elif lit is None:
                insn_name = 'push_nil'
                args = ()

        self.stream.append((insn_name, args))
        self.ip += len(args)+1

    def def_local(self, name):
        "Define a local variable."
        return self.env.alloc_local(name)

    def def_label(self, name):
        "Define a label at current ip."
        if self.labels.get(name) is not None:
            raise TypeError, "Duplicated label: %s" % name
        self.labels[name] = self.ip

    def emit_local(self, action, name, dyn_env=None):
        """\
        Emit an instruction to push or set local variable. The local variable
        is automatically searched in the current context and parents.
        """
        if dyn_env is None:
            env = self.env
            dyn = ""
        else:
            env = dyn_env
            dyn = "dynamic_"

        depth, idx = self.find_local_depth(name, env)
        if depth is None:
            raise UnboundVariable(name, "Unbound variable %s" % name)
        if depth == 0:
            postfix = ''
            args = (idx,)
        else:
            postfix = '_depth'
            args = (depth, idx)
        self.emit('%s%s_local%s' % (dyn, action, postfix), *args)

    def push_proc(self, args=[], rest_arg=False, parent_env=None):
        """\
        Return a builder for building a procedure. The returned builder
        is used to build the body of the procedure.

        Later when self.generate is called, builder.generate will be called
        automatically to get the procedure object, add it to the literals
        and push to the operand stack.
        """
        if parent_env is None:
            parent_env = self.env
        # create a new environment for procedure
        env = Environment(parent_env)
        # Define procedure arguments as local variables
        for x in args:
            env.alloc_local(x)

        bdr = Builder(env, result_t=Procedure)
        # Those properties are recorded in the builder and used
        # to construct the procedure later
        bdr.args = args
        bdr.rest_arg = rest_arg

        # generate_proc is a pseudo instruction
        self.stream.append(('generate_proc', bdr))
        self.ip += 3 # push_literal + fix_lexical (2+1)
        
        return bdr

    def generate(self):
        """\
        Generate a form with emitted instructions.
        """
        bc = array('i')
        for insn_name, args in self.stream:
            # pseudo instructions
            if insn_name == 'generate_proc':
                idx = len(self.literals)
                self.literals.append(args.generate())
                bc.append(INSN_MAP['push_literal'].opcode)
                bc.append(idx)
            # real instructions
            else:
                insn = INSN_MAP[insn_name]
                bc.append(insn.opcode)
                
                if insn_name in ['goto', 'goto_if_false', 'goto_if_not_false']:
                    bc.append(self.labels[args[0]])
                elif insn_name == 'push_literal':
                    bc.append(self.get_literal_idx(args[0]))
                else:
                    for x in args:
                        bc.append(x)

        return self.result_t(self, bc)

        
    ########################################
    # Helpers used internally
    ########################################
    def find_local_depth(self, name, env):
        """\
        Find the depth and index of a local variable. If no variable
        with the given name is found, return (None, None).
        """
        depth = 0
        while env is not None:
            idx = env.find_local(name)
            if idx is not None:
                return (depth, idx)
            depth += 1
            env = env.parent
        return (None, None)

    def get_literal_idx(self, lit):
        """\
        Return the index in literals list if there. Or else append
        the literal to the literals list.
        """
        for i in range(len(self.literals)):
            # make sure type is the same so that 42 and 42.0 will be
            # different literals
            l = self.literals[i]
            if type(l) is type(lit) and \
               l == lit:
                return i
        self.literals.append(lit)
        return len(self.literals)-1



import math

from .types.symbol import Symbol as sym
from .types.pair   import Pair as pair
from .proc         import Procedure
from .errors       import WrongArgNumber
from .errors       import WrongArgType
from .errors       import MiscError

class Primitive(object):
    """\
    Base class for all skime primitives.

    When primitive is called with arguments (1, 2, 3), the arity is first checked
    by calling prim.check_arity(3). Then the vm object is inserted as the first
    argument and the primitive called: prim.call(vm, 1, 2, 3). The vm is always
    the first argument of all primitives, but not count as argc.
    """
    def check_arity(self, argc):
        "Check whether this primitive is OK to execute with argc arguments."
        raise TypeError("check_arity is not implemented in abstract class Primitive")

    def call(self, *args):
        "Call the primitive with args."
        raise TypeError("call is not implemented in abstract class Primitive")


class PyPrimitive(Primitive):
    "Primitive wrapping a Python callable."
    def __init__(self, proc, arity):
        """\
        Create a PyPrimitive.

          proc  should be a Python callable.
          arity can be a tuple specifying the min and max number of arguments.
                either min or max or both can be -1, which means it is of no
                bound.
        """
        self.proc = proc
        self.arity = arity
        
    def check_arity(self, argc):
        min, max = self.arity
        if min > 0 and argc < min:
            raise WrongArgNumber("%s expects at least %d arguments, but got %d",
                                 (self.proc.__name__, min, argc))
        if max > 0 and argc > max:
            raise WrongArgNumber("%s expects at most %d arguments, but got %d",
                                 (self.proc.__name__, max, argc))

    def call(self, *args):
        return self.proc(*args)

    def __str__(self):
        return "<skime primitive => %s>" % self.proc.__name__

class PyCallable(Primitive):
    def __init__(self, proc):
        self.proc = proc

    def check_arity(self, argc):
        return True

    def call(self, vm, *args):
        return self.proc(*args)

def load_primitives(env):
    "Load primitives into an Environment."
    env.alloc_local('+', PyPrimitive(plus, (-1, -1)))
    env.alloc_local('-', PyPrimitive(minus, (1, -1)))
    env.alloc_local('*', PyPrimitive(mul, (-1, -1)))
    env.alloc_local('/', PyPrimitive(div, (1, -1)))
    env.alloc_local('=', PyPrimitive(equal, (-1, -1)))
    env.alloc_local('<', PyPrimitive(less, (2, -1)))
    env.alloc_local('>', PyPrimitive(more, (2, -1)))
    env.alloc_local('<=', PyPrimitive(less_equal, (2, -1)))
    env.alloc_local('>=', PyPrimitive(more_equal, (2, -1)))

    env.alloc_local('equal?', PyPrimitive(prim_equal, (2, 2)))
    env.alloc_local('eq?', PyPrimitive(prim_eqv, (2, 2)))
    env.alloc_local('eqv?', PyPrimitive(prim_eqv, (2, 2)))

    env.alloc_local("log", PyPrimitive(prim_log, (1, 1)))
    env.alloc_local("exp", PyPrimitive(prim_exp, (1, 1)))
    env.alloc_local("sin", PyPrimitive(prim_sin, (1, 1)))
    env.alloc_local("cos", PyPrimitive(prim_cos, (1, 1)))
    env.alloc_local("tan", PyPrimitive(prim_tan, (1, 1)))
    env.alloc_local("abs", PyPrimitive(prim_abs, (1, 1)))
    
    env.alloc_local('not', PyPrimitive(prim_not, (1, 1)))
        
    env.alloc_local('first', PyPrimitive(prim_first, (1, 1)))
    env.alloc_local('rest', PyPrimitive(prim_rest, (1, 1)))
    env.alloc_local('pair', PyPrimitive(prim_pair, (2, 2)))
    env.alloc_local('car', PyPrimitive(prim_first, (1, 1)))
    env.alloc_local('cdr', PyPrimitive(prim_rest, (1, 1)))
    env.alloc_local('cons', PyPrimitive(prim_pair, (2, 2)))
    env.alloc_local('set-first!', PyPrimitive(prim_set_first_x, (2, 2)))
    env.alloc_local('set-car!', PyPrimitive(prim_set_first_x, (2, 2)))
    env.alloc_local('set-rest!', PyPrimitive(prim_set_rest_x, (2, 2)))
    env.alloc_local('set-cdr!', PyPrimitive(prim_set_rest_x, (2, 2)))
    

    env.alloc_local('list', PyPrimitive(prim_list, (-1, -1)))

    for t,name in [(bool, "boolean?"),
                   (pair, "pair?"),
                   (sym, "symbol?"),
                   (str, "string?"),
                   ((int, long, float, complex), "number?"),
                   ((int, long, float), "rational?"),
                   ((int, long, float), "real?"),
                   ((int, long, float, complex), "complex?"),
                   ((int, long), "integer?"),
                   ((Procedure, Primitive), "procedure?")]:
        env.alloc_local(name, PyPrimitive(make_type_predict(t), (1, 1)))

    env.alloc_local('exact?', PyPrimitive(prim_exact_p, (1, 1)))
    env.alloc_local('inexact?', PyPrimitive(prim_inexact_p, (1, 1)))
    env.alloc_local('zero?', PyPrimitive(prim_zero_p, (1, 1)))
    env.alloc_local('positive?', PyPrimitive(prim_positive_p, (1, 1)))
    env.alloc_local('negative?', PyPrimitive(prim_negative_p, (1, 1)))
    env.alloc_local('even?', PyPrimitive(prim_even_p, (1, 1)))
    env.alloc_local('odd?', PyPrimitive(prim_odd_p, (1, 1)))
    env.alloc_local('max', PyPrimitive(prim_max, (1, -1)))
    env.alloc_local('min', PyPrimitive(prim_min, (1, -1)))
    env.alloc_local('quotient', PyPrimitive(prim_quotient, (2, 2)))
    env.alloc_local('modulo', PyPrimitive(prim_modulo, (2, 2)))
    env.alloc_local('remainder', PyPrimitive(prim_remainder, (2, 2)))
    env.alloc_local('gcd', PyPrimitive(prim_gcd, (-1, -1)))
    env.alloc_local('lcm', PyPrimitive(prim_lcm, (-1, -1)))
    env.alloc_local('floor', PyPrimitive(prim_floor, (1, 1)))
    env.alloc_local('ceiling', PyPrimitive(prim_ceiling, (1, 1)))
    env.alloc_local('truncate', PyPrimitive(prim_truncate, (1, 1)))
    env.alloc_local('round', PyPrimitive(prim_round, (1, 1)))
    env.alloc_local('asin', PyPrimitive(prim_asin, (1, 1)))
    env.alloc_local('acos', PyPrimitive(prim_acos, (1, 1)))
    env.alloc_local('atan', PyPrimitive(prim_atan, (1, 2)))
    env.alloc_local('sqrt', PyPrimitive(prim_sqrt, (1, 1)))
    env.alloc_local('expt', PyPrimitive(prim_expt, (2, 2)))
    env.alloc_local('null?', PyPrimitive(prim_null_p, (1, 1)))
    env.alloc_local('list?', PyPrimitive(prim_list_p, (1, 1)))

    env.alloc_local('apply', PyPrimitive(prim_apply, (1, -1)))
    env.alloc_local('map', PyPrimitive(prim_map, (2, -1)))

    env.alloc_local('string->symbol', PyPrimitive(prim_string_to_symbol, (1, 1)))
    env.alloc_local('symbol->string', PyPrimitive(prim_symbol_to_string, (1, 1)))
    env.alloc_local('number->string', PyPrimitive(prim_number_to_string, (1, 2)))
    env.alloc_local('string->number', PyPrimitive(prim_string_to_number, (1, 2)))
    env.alloc_local('string-append', PyPrimitive(prim_string_append, (-1, -1)))

def type_error_decorator(meth):
    "Decorate method to catch Python TypeError and raise skime WrongArgType"
    def new_meth(*args):
        try:
            return meth(*args)
        except TypeError, e:
            raise WrongArgType(e.message)
    return new_meth

@type_error_decorator
def plus(vm, *args):
    return sum(args)

@type_error_decorator
def mul(vm, *args):
    res = 1
    for x in args:
        res *= x
    return res

@type_error_decorator
def minus(vm, num, *args):
    if len(args) == 0:
        return -num
    for x in args:
        num -= x
    return num

@type_error_decorator
def div(vm, num, *args):
    if len(args) == 0:
        return 1.0/num
    if isinstance(num, (int, long)):
        num = float(num)
    for x in args:
        num /= x
    return num

def equal(vm, *args):
    if len(args) < 2:
        return True
    a = args[0]
    b = args[1]
    type_check(a, (int, long, float, complex))
    type_check(b, (int, long, float, complex))
    if a != b:
        return False
    for x in args[2:]:
        type_check(a, (int, long, float, complex))
        if x != a:
            return False
    return True

def less(vm, a, b, *args):
    if a >= b:
        return False
    for x in args:
        if b >= x:
            return False
        b = x
    return True

def more(vm, a, b, *args):
    if a <= b:
        return False
    for x in args:
        if b <= x:
            return False
        b = x
    return True

def less_equal(vm, a, b, *args):
    if a > b:
        return False
    for x in args:
        if b > x:
            return False
        b = x
    return True

def more_equal(vm, a, b, *args):
    if a < b:
        return False
    for x in args:
        if b < x:
            return False
        b = x
    return True

def prim_positive_p(vm, arg):
    type_check(arg, (int, long, float))
    return arg > 0
def prim_negative_p(vm, arg):
    type_check(arg, (int, long, float))
    return arg < 0
def prim_odd_p(vm, arg):
    type_check(arg, (int, long))
    return arg % 2 != 0
def prim_even_p(vm, arg):
    type_check(arg, (int, long))
    return arg % 2 == 0
def prim_max(vm, *args):
    return max(args)
def prim_min(vm, *args):
    return min(args)
def prim_quotient(vm, a, b):
    type_check(a, (int, long))
    type_check(b, (int, long))
    return a/b
# remainder has the same sign as b
def prim_remainder(vm, a, b):
    type_check(a, (int, long))
    type_check(b, (int, long))
    if a > 0:
        return a%b
    return (-a)%(-b)
# modulo has the same sign as a
def prim_modulo(vm, a, b):
    type_check(a, (int, long))
    type_check(b, (int, long))
    return a%b


def gcd(a, b):
    while b != 0:
        b, a = a % b, b
    return a
def lcm(a, b):
    return a*b/gcd(a, b)
def prim_gcd(vm, *args):
    if args is None:
        return 0
    if len(args) == 1:
        return abs(args[0])
    a, b, args = args[0], args[1], args[2:]
    type_check(a, (int, long))
    type_check(b, (int, long))
    g = gcd(a, b)
    for x in args:
        type_check(x, (int, long))
        g = gcd(g, x)
    return abs(g)
    
def prim_lcm(vm, *args):
    if args is None:
        return 1
    if len(args) == 1:
        return abs(args[0])
    a, b, args = args[0], args[1], args[2:]
    type_check(a, (int, long))
    type_check(b, (int, long))
    l = lcm(a, b)
    for x in args:
        type_check(x, (int, long))
        l = lcm(l, x)
    return abs(l)

    
@type_error_decorator
def prim_floor(vm, a):
    return math.floor(a)
@type_error_decorator
def prim_ceiling(vm, a):
    return math.ceil(a)
@type_error_decorator
def prim_truncate(vm, a):
    if a > 0:
        return math.floor(a)
    return math.ceil(a)
@type_error_decorator
def prim_round(vm, a):
    return round(a)

@type_error_decorator
def prim_exp(vm, arg):
    return math.exp(arg)

@type_error_decorator
def prim_log(vm, arg):
    return math.log(arg)

@type_error_decorator
def prim_sin(vm, arg):
    return math.sin(arg)

@type_error_decorator
def prim_cos(vm, arg):
    return math.cos(arg)

@type_error_decorator
def prim_tan(vm, arg):
    return math.tan(arg)

@type_error_decorator
def prim_asin(vm, arg):
    return math.asin(arg)

@type_error_decorator
def prim_acos(vm, arg):
    return math.acos(arg)

@type_error_decorator
def prim_atan(vm, arg, *arg2):
    if arg2 is None:
        return math.atan(arg)
    else:
        return math.atan(float(arg2[0])/arg)

@type_error_decorator
def prim_sqrt(vm, arg):
    return math.sqrt(arg)

@type_error_decorator
def prim_expt(vm, a, b):
    return a ** b

@type_error_decorator
def prim_abs(vm, arg):
    return abs(arg)

def prim_not(vm, arg):
    if arg is False:
        return True
    return False

def prim_first(vm, arg):
    type_check(arg, pair)
    return arg.first
def prim_rest(vm, arg):
    type_check(arg, pair)
    return arg.rest
def prim_pair(vm, a, b):
    return pair(a, b)
def prim_set_first_x(vm, arg, val):
    type_check(arg, pair)
    arg.first = val
def prim_set_rest_x(vm, arg, val):
    type_check(arg, pair)
    arg.rest = val

def prim_exact_p(vm, arg):
    if isinstance(arg, (int, long)):
        return True
    # python complex are always inexact
    if isinstance(arg, (float, complex)):
        return False
    raise WrongArgType("Expecting a number, but got %s" % arg)
def prim_inexact_p(vm, arg):
    return not prim_exact_p(vm, arg)

def prim_zero_p(vm, arg):
    type_check(arg, (int, long, complex, float))
    return arg == 0
    
def prim_null_p(vm, arg):
    return arg is None

# list?, detect circular list
def prim_list_p(vm, val):
    obj1 = val
    obj2 = val
    while True:
        if obj1 is None:
            return True
        if not isinstance(obj1, pair):
            return False

        obj1 = obj1.rest
        if obj1 is None:
            return True
        if not isinstance(obj1, pair):
            return False

        obj1 = obj1.rest
        obj2 = obj2.rest

        # circular
        if obj1 is obj2:
            break
        
    return False

def prim_list(vm, *args):
    lst = None
    for x in reversed(args):
        lst = pair(x, lst)
    return lst


def prim_apply(vm, proc, *args):
    if len(args) == 0:
        return vm.apply(proc, args)
    argv = list(args[:-1])
    arglst = args[-1]
    while isinstance(arglst, pair):
        argv.append(arglst.first)
        arglst = arglst.rest
    if arglst is not None:
        raise WrongArgType("The last argument of apply should be a valid list, but got %s" % args[-1])
    return vm.apply(proc, argv)

def prim_map(vm, proc, *lists):
    res = []
    while True:
        args = []
        end = False
        lists = list(lists)
        for i in range(len(lists)):
            lst = lists[i]
            if not isinstance(lst, pair):
                if lst is None:
                    end = True
                else:
                    raise WrongArgType("Arguments of map should be valid lists.")
            else:
                if end:
                    raise MiscError("Lists supplied to map should be all of the same length.")
                args.append(lst.first)
                lists[i] = lst.rest
        if end:
            break
        res.append(vm.apply(proc, args))
    rest = None
    for x in reversed(res):
        rest = pair(x, rest)
    return rest

def prim_string_to_symbol(vm, name):
    type_check(name, str)
    return sym(name)
def prim_symbol_to_string(vm, s):
    type_check(s, sym)
    return s.name

def prim_number_to_string(vm, num, radix=10):
    if radix == 10:
        return str(num)
    type_check(num, (int, long))
    minus = False
    if num < 0:
        minus = True
        num = -num
    
    if radix == 2:
        ditigs = []
        while num != 0:
            digits.append(num & 1)
            num = num >> 1
        num.reverse()
        fmt = ''.join(num)

    if radix == 8:
        fmt = '%o' % num
    if radix == 16:
        fmt = '%X' % num

    if minus:
        return '-' + fmt
    return fmt

def prim_string_to_number(vm, s, radix=10):
    try:
        return int(s, radix)
    except ValueError:
        if radix != 10:
            raise MiscError("Only radix 10 is permitted for decimal number")
        try:
            if s[-1] == 'i':
                # complex number
                return complex(s[:-1]+'j')
            else:
                return float(s)
        except ValueError:
            return False

@type_error_decorator
def prim_string_append(vm, *strings):
    return ''.join(strings)

def prim_equal(vm, a, b):
    return a == b

def prim_eqv(vm, a, b):
    return a is b


########################################
# Helper for primitives
########################################
def make_type_predict(tt):
    def predict(vm, obj):
        return isinstance(obj, tt)
    return predict

def type_check(obj, t):
    if not isinstance(obj, t):
        raise WrongArgType("Expecting type %s, but got %s (type %s)" % \
                           (t, obj, type(obj)))

def iter_list(lst, excp_t=WrongArgType):
    while isinstance(lst, pair):
        yield lst.first
        lst = lst.rest
    if lst is not None:
        raise excp_t("Not a proper list")



class Pair(object):
    """\
    The Pair pair of Scheme.

    Pair can be chained to form a list. The Python value None acts as
    end-of-list.

      Pair(1, Pair(2, None)) <==> '(1 2) <==> '(1 . (2 . ()))
    """
    __slots__ = ['first', 'rest']

    def __init__(self, first, rest):
        self.first = first
        self.rest = rest

    def get_car(self):
        return self.first
    def set_car(self, val):
        self.first = val
    def get_cdr(self):
        return self.rest
    def set_cdr(self, val):
        self.rest = val
    car = property(get_car, set_car)
    cdr = property(get_cdr, set_cdr)

    def __eq__(self, other):
        return isinstance(other, Pair) and \
               self.first == other.first and \
               self.rest == other.rest
    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        segments = [self.first.__str__()]
        elems = self.rest

        while isinstance(elems, Pair):
            segments.append(elems.first.__str__())
            elems = elems.rest

        if elems is not None:
            segments.append(".")
            segments.append(elems.__str__())
        return '(' + ' '.join(segments) + ')'



class Location(object):
    """\
    A location of a variable, including the Environment object and
    the index in it.
    """
    __slots__ = ('env', 'idx')
    
    def __init__(self, env, idx):
        self.env = env
        self.idx = idx

    def __eq__(self, other):
        "Test if two Location object refer to the same location."
        return isinstance(other, Location) and \
               self.idx == other.idx and \
               self.env is other.env

    # Python rocks! Just Keep It Simple and verboSe.
    # When you defined __eq__, please also define __ne__ or
    # else != will compare by the memory address of your
    # objects
    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return '<Location idx=%s, env=%s>' % (self.idx, self.env)

class Undef(object):
    undef = None
    def __new__(cls):
        if cls.undef is None:
            cls.undef = object.__init__(cls)
        return cls.undef
        
    def __repr__(self):
        return '<undef>'
    def __str__(self):
        return '<undef>'

class Environment(object):
    """\
    An environment object holds the local variables of a scope
    and is chained through the lexical scope.
    """

    def __init__(self, parent=None):
        # The lexical parent
        self.parent = parent

        if parent is not None:
            self.vm = parent.vm
        else:
            self.vm = None

        # The values of local variables
        self.locals = []
        # The names of local variables
        self.locals_name = []
        # The mapping from name to index
        self.locals_map = {}

    def dup(self):
        """\
        Create a copy of self.
        """
        env = Environment(self.parent)
        env.locals = list(self.locals)
        env.locals_name = list(self.locals_name)
        env.locals_map = dict(self.locals_map)
        return env

    def assign_local(self, idx, value):
        """\
        Assign value to the local variable stored at idx.
        """
        self.locals[idx] = value

    def read_local(self, idx):
        """\
        Get the value of the local variable stored at idx.
        """
        return self.locals[idx]

    def alloc_local(self, name, value=Undef()):
        """\
        Allocate space for storing local variable. Return
        the index for the variable.

        Allocating a variable with the same name several times
        is OK. They will be stored at the same location, and
        value assigned later will overwrite earlier values.
        """
        idx = self.locals_map.get(name)
        if idx is not None:
            if value is not Undef():
                self.locals[idx] = value
            return idx
        idx = len(self.locals)
        self.locals_name.append(name)
        self.locals.append(value)
        self.locals_map[name] = idx
        return idx

    def find_local(self, name):
        """\
        Find the location(index) where the local variable is
        stored. Return None if no variable with given name is
        found.
        """
        return self.locals_map.get(name)

    def get_name(self, idx):
        """\
        Get the name of the local variable stored at the given
        location(index). Used in debugging.
        """
        return self.locals_name[idx]

    def lookup_location(self, name):
        """\
        Find the location of variable with given name. Recursively
        find in parent when necessary.
        """
        env = self
        while env is not None:
            idx = env.find_local(name)
            if idx is not None:
                return Location(env, idx)
            env = env.parent
        return None

    def __repr__(self):
        return "<Environment @%X>" % id(self)



# A Scheme expression is compiled into a Form. The Form object hold the
# bytecode of the expression. To evaluate the form, a new Context is set up.
# Instruction pointer and operand stack are held in the Context object.
# Local variables are held in an Environment object, which are chained
# through the lexical scope.

import os.path

from .ctx               import Context
from .env               import Environment
from .                  import insns
from .types.pair        import Pair
from .proc              import Procedure
from .prim              import Primitive, load_primitives
from .insns             import run
from .types.pair        import Pair as pair

from .compiler.parser   import parse
from .compiler.compiler import Compiler

from .errors            import WrongArgType

class VM(object):

    def __init__(self):
        self.compiler = Compiler()
        
        self.env = Environment()
        self.env.vm = self
        load_primitives(self.env)

        self.ctx = Context(None, self.env, None)

        self.load(os.path.join(os.path.dirname(__file__),
                               'scheme',
                               'prim.scm'))

    def run(self, form):
        return form.eval(self.env, self)

    def load(self, path):
        io = open(path)
        content = io.read()
        io.close()

        return self.eval_string("(begin %s)" % content)

    def eval_string(self, script):
        return self.run(self.compiler.compile(parse(script), self.env))

    def apply(self, proc, args):
        if isinstance(proc, Procedure):
            proc.check_arity(len(args))

            ctx = Context(proc, proc.env.dup(), self.ctx)
            for i in range(proc.fixed_argc):
                ctx.env.assign_local(i, args[i])
            if proc.fixed_argc != proc.argc:
                rest = None
                for i in range(len(args)-1, proc.fixed_argc-1, -1):
                    rest = Pair(args[i], rest)
                ctx.env.assign_local(proc.fixed_argc, rest)

            return run(ctx)
        
        elif isinstance(proc, Primitive):
            proc.check_arity(len(args))
            return proc.call(self, *args)
        
        else:
            raise WrongArgType("Not a skime callable: %s" % proc)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['skime', 'skime.compiler', 'skime.types']

package_data = \
{'': ['*'], 'skime': ['scheme/*']}

install_requires = \
['PyYAML>=5,<6']

setup_kwargs = {
    'name': 'skime',
    'version': '0.1.0',
    'description': '',
    'long_description': None,
    'author': None,
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=2.7,<2.8',
}


setup(**setup_kwargs)



class Error(Exception):
    "Base class for all exceptions in skime"

class CompileError(Error):
    "Compile errors"

class UnboundVariable(CompileError):
    def __init__(self, name, message):
        Exception.__init__(self, message)
        self.name = name

class WrongArgNumber(CompileError):
    pass

class SyntaxError(CompileError):
    pass

class ParseError(CompileError):
    pass

class MiscError(Error):
    pass

class WrongArgType(Error):
    pass



from .types.pair   import Pair as pair
from .types.symbol import Symbol as sym
from .errors       import SyntaxError

class Macro(object):
    def __init__(self, env, body):
        self.lexical_parent = env
        try:
            # Process literals
            literals = body.first
            lit = []
            while isinstance(literals, pair):
                lit.append(literals.first)
                literals = literals.rest
            if literals is not None:
                raise SyntaxError("Invalid syntax rule format: literals should be a proper list.")

            self.rules = []
            # Process syntax rules
            rules = body.rest
            while rules is not None:
                rule = rules.first
                self.rules.append(SyntaxRule(rule, lit, env))
                rules = rules.rest
        except AttributeError:
            raise SyntaxError("Invalid syntax for syntax-rules form")

    def transform(self, env, form):
        for rule in self.rules:
            try:
                md = rule.match(env, form)
                return rule.expand(env, md)
            except MatchError:
                pass
        raise SyntaxError("Can not find syntax rule to match the form %s" % form)

class SyntaxRule(object):
    def __init__(self, rule, literals, env):
        if not isinstance(rule, pair) or not isinstance(rule.rest, pair):
            raise SyntaxError("Expecting (pattern template) for syntax rule, but got %s" % rule)
        if rule.rest.rest is not None:
            raise SyntaxError("Extra expressions in syntax rule: %s" % rule)
        self.env = env
        self.variables = {}
        self.matcher = self.compile_pattern(rule.first, literals)
        self.template = self.compile_template(rule.rest.first)

    def match(self, env, form):
        md = MatchDict()
        if not isinstance(form, pair):
            raise SyntaxError("Invalid macro matching against the form %s" % form)
        # skip the first element, which is the macro keyword
        self.matcher.match(env, pair(form.rest, None), md)
        return md

    def expand(self, env, md):
        dc_factory = ClosureFactory(env)
        expr = self.template.expand(dc_factory, md, [])[0]
        return expr, dc_factory.closures

    ########################################
    # Pattern compiling
    ########################################
    def compile_pattern(self, pattern, literals):
        """\
        Compile pattern into a matcher.
        """
        if not isinstance(pattern, pair):
            raise SyntaxError("Invalid pattern for macro: %s" % pattern)
        # skip the first element, it should be the macro keyword or underscope
        # NOTE, if some invalid expressions are put here, they will be silently
        # ignored
        pattern = pattern.rest

        return self._compile_pattern(pattern, literals)

    def _compile_pattern(self, pat, literals):
        if isinstance(pat, pair):
            mt = SequenceMatcher()
            while isinstance(pat, pair):
                submt = self._compile_pattern(pat.first, literals)
                mt.add_matcher(submt)
                pat = pat.rest
                if isinstance(pat, pair) and pat.first == sym('...'):
                    submt.ellipsis = True
                    pat = pat.rest
            if pat is not None:
                submt = self._compile_pattern(pat, literals)
                mt.add_matcher(RestMatcher(submt))
            return mt

        if isinstance(pat, sym):
            if pat in literals:
                return LiteralMatcher(self.env, pat.name)
            if pat == sym('_'):
                return UnderscopeMatcher()
            if self.variables.get(pat.name) is not None:
                raise SyntaxError("Duplicated variable in macro: %s" % pat)
            self.variables[pat.name] = True
            return VariableMatcher(pat.name)

        return ConstantMatcher(pat)

    ########################################
    # Template compiling
    ########################################
    def compile_template(self, expr):
        if isinstance(expr, pair):
            tmpl = SequenceTemplate()
            while isinstance(expr, pair):
                sub_tmpl = self.compile_template(expr.first)
                expr = expr.rest
                while isinstance(expr, pair) and expr.first == sym('...'):
                    sub_tmpl.nflatten += 1
                    expr = expr.rest
                tmpl.add_tmpl(sub_tmpl)
            if expr is not None:
                sub_tmpl = self.compile_template(expr)
                tmpl.set_tail(sub_tmpl)
            return tmpl
        if isinstance(expr, sym) and self.variables.get(expr.name) is not None:
            return VariableTemplate(expr.name)
        return ConstantTemplate(expr)
            

########################################
# Pattern matching
########################################
class MatchError(Exception):
    pass

class MatchDict(dict):
    """\
    A MatchDict hold the matched value of variable patterns. It is an error to
    have duplicated variable patterns with the same name.
    """
    def __setitem__(self, key, value):
        dict.__setitem__(self, key, value)
    def __str__(self):
        return "<MatchDict %s>" % dict.__str__(self)

class Ellipsis(list):
    """\
    Ellipsis holds the zero or more value of an ellipsis pattern.
    """
    def __init__(self, *value):
        list.__init__(self, value)
    def __repr__(self):
        return "<Ellipsis %s>" % list.__repr__(self)
            

class EllipsisMatchDict(dict):
    """\
    Like MatchDict, excpet that all values are Ellipsis.
    """
    def __setitem__(self, key, value):
        el = self.get(key)
        if el is None:
            # no such key yet
            dict.__setitem__(self, key, Ellipsis(value))
        else:
            # already has the key
            el.append(value)
    def __str__(self):
        return "<EllipsisMatchDict %s>" % dict.__str__(self)
            

class Matcher(object):
    "The base class for all matchers."
    def __init__(self, name):
        self.ellipsis = False
        self.name = name

    def class_name(self):
        return "%s%s" % (self.__class__.__name__,
                         self.ellipsis and "*" or "")

    def match(self, env, expr, match_dict):
        """\
        Match against expr and return the remaining expression.
        If can not match, raise MatchError.
        """
        raise MatchError("%s: match not implemented" % self)

    def __str__(self):
        return '<%s name=%s>' % (self.class_name(), self.name)

class LiteralMatcher(Matcher):
    """\
    A LiteralMatcher only matches the symbol with the same lexical binding
    or both has no binding.
    """
    def __init__(self, env, name):
        Matcher.__init__(self, name)
        self.loc = self.get_loc(env, name)

    def match(self, env, expr, match_dict):
        if self.ellipsis:
            try:
	    	while True:
		    self.match_literal(env, expr)
                    expr = expr.rest
            except MatchError:
                pass
            return expr

        self.match_literal(env, expr)   # may raise MatchError
        return expr.rest

    def get_loc(self, env, name):
        if env is None:
            return None
        return env.lookup_location(name)

    def match_literal(self, env, expr):
        if not isinstance(expr, pair) or \
           not isinstance(expr.first, sym):
            raise MatchError("%s: can not match %s" % (self, expr))
        loc = self.get_loc(env, expr.first.name)
        if self.loc != loc:
            raise MatchError("%s: can not match %s with different lexical binding" % (self, expr.first.name))

class ConstantMatcher(Matcher):
    """\
    Matches any constant.
    """
    def __init__(self, value):
        Matcher.__init__(self, None)
        self.value = value
    def match(self, env, expr, match_dict):
        if self.ellipsis:
            while isinstance(expr, pair) and \
                  expr.first == self.value:
                expr = expr.rest
            return expr
        
        if not isinstance(expr, pair) or \
           expr.first != self.value:
            raise MatchError("%s: can not match %s" % (self, expr))
        return expr.rest
    def __str__(self):
        return "<%s value=%s>" % (self.class_name(), self.value)

class VariableMatcher(Matcher):
    """\
    A variable match any single expression.
    """
    def match(self, env, expr, match_dict):
        if self.ellipsis:
            result = Ellipsis()
            while isinstance(expr, pair):
                result.append(expr.first)
                expr = expr.rest
        else:
            if not isinstance(expr, pair):
                raise MatchError("%s: unable to match %s" % (self, expr))
            result = expr.first
            expr = expr.rest

        match_dict[self.name] = result
        return expr
        
class UnderscopeMatcher(Matcher):
    """\
    An underscope match any single expression and discard the matched result.
    """
    def __init__(self):
        Matcher.__init__(self, '_')
    def match(self, env, expr, match_dict):
        if self.ellipsis:
            while isinstance(expr, pair):
                expr = expr.rest
        else:
            if not isinstance(expr, pair):
                raise MatchError("%s: unable to match %s" % (self, expr))
            expr = expr.rest

        return expr
    
class RestMatcher(Matcher):
    """\
    RestMatcher match against the rest of a list like (a b . c), where c will
    be a RestMatcher. The matcher is implemented by wrapping another matcher.
    """
    def __init__(self, matcher):
        self.matcher = matcher
    def match(self, env, expr, match_dict):
        return self.matcher.match(env, pair(expr, None), match_dict)
    def __str__(self):
        return "<RestMatcher: matcher=%s>" % self.matcher

class SequenceMatcher(Matcher):
    """\
    SequenceMatcher hold a sequence of matchers and matches them in that order.
    """
    def __init__(self):
        Matcher.__init__(self, None)
        self.sequence = []

    def match(self, env, expr, match_dict):
        if self.ellipsis:
            md = EllipsisMatchDict()
            try:
                while isinstance(expr, pair):
                    self.match_sequence(env, expr, md)
                    expr = expr.rest
            except MatchError:
                pass
            # merge data in EllipsisMatchDict into match_dict
            match_dict.update(md)
            return expr
        else:
            if not isinstance(expr, pair):
                raise MatchError("%s: unable to match %s" % (self, expr))
            self.match_sequence(env, expr, match_dict)
            return expr.rest

    def match_sequence(self, env, expr, match_dict):
        expr = expr.first
        for m in self.sequence:
            expr = m.match(env, expr, match_dict)
        if expr is not None:
            # not matched expressions
            raise MatchError("%s: ramaining expressions not matched: %s" % (self, expr))
        

    def add_matcher(self, matcher):
        self.sequence.append(matcher)

    def __str__(self):
        return "<%s sequence=[%s]>" % (self.class_name(),
                                       ', '.join([m.__str__()
                                                  for m in self.sequence]))

########################################
# Template expanding
########################################
class DynamicClosure(object):
    """\
    A macro is evaluated in the lexical scope of the macro. However,
    expressions captured by pattern variables should be evaluated
    in the scope where the macro is expanded. This object is used
    to wrap the expression and carry with it its dynamic scope.

    slots are:
     - lexical_parent: the environment where the expression belongs to.
     - expression: the expression wrapped.
     - form: the compiled form of the expression.
    """
    __slots__ = ('lexical_parent', 'expression', 'form')
    
    def __init__(self, env, expr):
        self.lexical_parent = env
        self.expression = expr

    def __eq__(self, o):
        return isinstance(o, DynamicClosure) and \
               self.lexical_parent == o.lexical_parent and \
               self.expression == o.expression
    def __ne__(self, o):
        return not self.__eq__(o)

    def __str__(self):
        return self.__repr__()
    def __repr__(self):
        return "<DynamicClosure expr=%s, env=%s>" % (self.expression,
                                                     self.lexical_parent)

class SymbolClosure(DynamicClosure):
    def __hash__(self):
        return self.expression.name.__hash__()

class ClosureFactory(object):
    "Create and hold DynamicClosure."
    def __init__(self, env):
        self.env = env
        self.values = []
        self.closures = []

    def make_closure(self, value):
        if isinstance(value, DynamicClosure):
            return value
        if isinstance(value, sym):
            sc = self.get_closure(sym)
            if sc is None:
                self.values.append(value)
                sc = SymbolClosure(self.env, value)
                self.closures.append(sc)
            return sc
        if isinstance(value, pair):
            dc = self.get_closure(sym)
            if dc is None:
                self.values.append(value)
                dc = DynamicClosure(self.env, value)
                self.closures.append(dc)
            return dc
        # other values are considered environment-indenpendent
        return value

    def get_closure(self, value):
        # the closures in a macro is generally not too many
        # so linear search is OK.
        for i in range(len(self.values)):
            if value == self.values[i]:
                return self.closures[i]
        return None

# There are the following kinds of templates:
#  - symbol:
#    - lexical binding: symbol referencing to the lexical binding where the macro is defined
#    - macro variable symbol: will be replaced by the matched value
#    - un-recognized symbol: will be renamed to avoid conflict
#  - pair: expand recursively
#  - other: expand as constant

class Template(object):
    "The base class for all template."
    def __init__(self):
        self.nflatten = 0
        
    def class_name(self):
        return self.__class__.__name__
    
    def expand(self, dc_factory, md, nflatten=0):
        "Expand the template under match dict md."
        raise SyntaxError("Attempt to expand an abstract template.")

class ConstantTemplate(Template):
    "Template that will expand to a constant."
    def __init__(self, value):
        Template.__init__(self)
        self.value = value

    def expand(self, dc_factory, md, idx=[]):
        return (self.value, )
    
    def __str__(self):
        return "<%s value=%s>" % (self.class_name(), self.value)


class VariableTemplate(Template):
    "Template that reference to a macro variable."
    def __init__(self, name):
        Template.__init__(self)
        self.name = name

    def expand(self, dc_factory, md, idx=[]):
        val = md.get(self.name, Ellipsis())
        for i in idx:
            val = val[i]
        nflatten = self.nflatten
        val = [val]
        while nflatten > 0:
            val = self.flatten(val)
            nflatten -= 1
        if len(val) > 0 and isinstance(val[0], Ellipsis):
            raise SyntaxError("Ellipsis after variable %s is less than expected." % self.name)
        return [dc_factory.make_closure(v) for v in val]

    def flatten(self, val):
        "Flatten ellipsis."
        res = []
        for x in val:
            if not isinstance(x, Ellipsis):
                raise SyntaxError("Too many ellipsis for variable %s" % self.name)
            res.extend(x)
        return res

    def __str__(self):
        return "<%s name=%s>" % (self.class_name(), self.name)

class SequenceTemplate(Template):
    "Template that aggregate a sequence of sub-templates."
    default_tail = ConstantTemplate(None)
    
    def __init__(self):
        Template.__init__(self)
        self.sequence = []
        self.tail = SequenceTemplate.default_tail

        self.ellipsis_names = []

    def add_tmpl(self, tmpl):
        self.calc_ellipsis_names(tmpl)
        self.sequence.append(tmpl)

    def set_tail(self, tmpl):
        """\
        Set the tail template of the sequence. It is normally
        default to ConstantTemplate(None).
        """
        self.calc_ellipsis_names(tmpl)
        self.tail = tmpl

    def calc_ellipsis_names(self, tmpl):
        if isinstance(tmpl, VariableTemplate):
            self.ellipsis_names.append(tmpl.name)
        elif isinstance(tmpl, SequenceTemplate):
            self.ellipsis_names.extend(tmpl.ellipsis_names)

    def expand(self, dc_factory, md, idx=[]):
        return self.expand_flatten(dc_factory, md, idx, self.nflatten)

    def expand_flatten(self, dc_factory, md, idx, flatten):
        if flatten == 0:
            return self.expand_0(dc_factory, md, idx)
        length = 0
        for name in self.ellipsis_names:
            var = md.get(name, Ellipsis())
            for i in idx:
                var = var[i]
            if not isinstance(var, Ellipsis):
                raise SyntaxError("Too many ellipsis after variable %s" % name)
            if length == 0 or length == len(var):
                length = len(var)
            else:
                raise SyntaxError("Incompatible ellipsis match counts for variable %s" % name)
        if length > 0:
            idx.append(0)
            res = []
            for i in range(length):
                idx[-1] = i
                res.extend(self.expand_flatten(dc_factory, md, idx, flatten-1))
            idx.pop()
            return res
        else:
            return ()

    def expand_0(self, dc_factory, md, idx):
        elems = []
        for tmpl in self.sequence:
            elems.extend(tmpl.expand(dc_factory, md, idx))
        rest = self.tail.expand(dc_factory, md, idx)[0]
        for elem in reversed(elems):
            rest = pair(elem, rest)
        return [rest]
        
    def __str__(self):
        return "<%s sequence=[%s], tail=%s>" % (self.class_name(),
                                                ', '.join([tmpl.__str__()
                                                           for tmpl in self.sequence]),
                                                self.tail)



from ..iset import INSTRUCTIONS


def disasm(io, form):
    bytecode = form.bytecode
    env = form.env
    literals = form.literals
    
    ip = 0
    while ip < len(bytecode):
        io.write("%04X " % ip)
        instr = INSTRUCTIONS[bytecode[ip]]
        io.write("%20s " % instr.name)
        if instr.name in ['push_local', 'set_local']:
            io.write('idx: %d' % bytecode[ip+1])
            io.write(', name: ')
            io.write(env.get_name(bytecode[ip+1]))
        elif instr.name in ['push_local_depth', 'set_local_depth']:
            depth = bytecode[ip+1]
            idx = bytecode[ip+2]
            io.write("depth=%d, idx=%d" % (depth, idx))
            penv = env
            while depth > 0:
                penv = penv.parent
                depth -= 1
            io.write(" (name: %s)" % penv.get_name(idx))
        elif instr.name in ['goto', 'goto_if_not_false', 'goto_if_false']:
            io.write("ip=0x%04X" % bytecode[ip+1])
        elif instr.name == 'dynamic_set_local':
            lit = bytecode[ip-1]
            io.write('idx: %d' % bytecode[ip+1])
            io.write(', name: ')
            io.write(literals[lit].expression.name)
        else:
            io.write(', '.join(["%s=%s" % (name, val)
                                for name, val in zip(instr.operands,
                                                     bytecode[ip+1:
                                                              ip+len(instr.operands)+1])]))
        io.write('\n')
        ip += instr.length
